using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using HtmlAgilityPack;
using SubtitleDownloader.Core;

namespace GammaJul.MediaPortal.Subtitles.SousTitresEu {

	/// <summary>
	/// Implementation of <see cref="SubtitleSearcher{TQuery}"/> for episode search.
	/// </summary>
	internal sealed class EpisodeSubtitleSearcher : SubtitleSearcher<EpisodeSearchQuery> {

		/// <summary>
		/// Gets the URL of the page containing the listing of series or movies.
		/// </summary>
		protected override string ListPageUrl {
			get { return SiteUrl + "series.html"; }
		}

		/// <summary>
		/// Gets the download base URL for links found onto <see cref="SubtitleSearcher{TQuery}.ListPageUrl"/>.
		/// </summary>
		protected override string DownloadBaseUrl {
			get { return SiteUrl + "series/"; }
		}

		/// <summary>
		/// Returns the title of the serie of movie queried.
		/// </summary>
		/// <returns>A title.</returns>
		protected override string GetTitleCore() {
			return Query.SerieTitle;
		}

		/// <summary>
		/// Returns subtitles nodes potentially matching the query from a serie or movie page.
		/// </summary>
		/// <param name="documentNode">The document node of the serie or movie page.</param>
		/// <returns>An enumeration of <see cref="HtmlNode"/> containing potential subtitles.</returns>
		protected override IEnumerable<HtmlNode> GetSubtitleNodes(HtmlNode documentNode) {
			string xpath = String.Format(CultureInfo.InvariantCulture, "//div[@id='saison-{0}']//a[@class='subList']", Query.Season);
			return documentNode.SelectNodes(xpath) ?? Enumerable.Empty<HtmlNode>();
		}

		/// <summary>
		/// Returns whether a subtitle node should be filtered from results.
		/// </summary>
		/// <param name="subtitleNode">The subtitle node to check.</param>
		/// <returns><c>true</c> if <paramref name="subtitleNode"/> should be filtered from results, <c>false</c> otherwise.</returns>
		protected override bool IsFiltered(HtmlNode subtitleNode) {
			HtmlNode numNode = subtitleNode.SelectSingleNode(String.Format(CultureInfo.InvariantCulture, "span[@class='episodenum']"));
			if (numNode == null)
				return true;

			string episodeNum = String.Format(CultureInfo.InvariantCulture, "{0}&times;{1:00}", Query.Season, Query.Episode);
			if (!episodeNum.Equals(numNode.InnerText, StringComparison.OrdinalIgnoreCase)
			&& !("S" + Query.Season).Equals(numNode.InnerText, StringComparison.Ordinal))
				return true;

			return false;
		}

		/// <summary>
		/// Initializes a new instance of <see cref="EpisodeSubtitleSearcher"/>.
		/// </summary>
		/// <param name="query">The query associated with the search.</param>
		/// <param name="timeout">The timeout of the operation.</param>
		internal EpisodeSubtitleSearcher(EpisodeSearchQuery query, TimeSpan timeout)
			: base(query, timeout) {
		}
	}

}