using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Web;
using HtmlAgilityPack;
using SubtitleDownloader.Core;

namespace GammaJul.MediaPortal.Subtitles.SousTitresEu {

	/// <summary>
	/// Implementation of <see cref="SubtitleSearcher{TQuery}"/> for movie search.
	/// </summary>
	internal sealed class MovieSubtitleSearcher : SubtitleSearcher<ImdbSearchQuery> {

		/// <summary>
		/// Gets the URL of the page containing the listing of series or movies.
		/// </summary>
		protected override string ListPageUrl {
			get { return SiteUrl + "films.html"; }
		}

		/// <summary>
		/// Gets the download base URL for links found onto <see cref="SubtitleSearcher{TQuery}.ListPageUrl"/>.
		/// </summary>
		protected override string DownloadBaseUrl {
			get { return SiteUrl + "films/"; }
		}

		/// <summary>
		/// Returns the title of the serie of movie queried.
		/// </summary>
		/// <returns>A title.</returns>
		protected override string GetTitleCore() {
			HtmlDocument document = LoadDocument(String.Format(CultureInfo.InvariantCulture, "http://www.imdb.com/title/tt{0}/", Query.ImdbId));
			HtmlNode headerNode = document.DocumentNode.SelectSingleNode("//td[@id='overview-top']/h1");
			if (headerNode == null)
				return null;

			string title = null;
			HtmlNode extraTitleContainerNode = headerNode.SelectSingleNode("span[@class='title-extra']");
			if (extraTitleContainerNode != null)
				title = GetFirstText(extraTitleContainerNode);
			if (String.IsNullOrEmpty(title))
				title = GetFirstText(headerNode);
			if (!String.IsNullOrEmpty(title))
				title = HttpUtility.HtmlDecode(title);
			return title;
		}

		/// <summary>
		/// Returns subtitles nodes potentially matching the query from a serie or movie page.
		/// </summary>
		/// <param name="documentNode">The document node of the serie or movie page.</param>
		/// <returns>An enumeration of <see cref="HtmlNode"/> containing potential subtitles.</returns>
		protected override IEnumerable<HtmlNode> GetSubtitleNodes(HtmlNode documentNode) {
			return documentNode.SelectNodes("//div[@id='film-content']//a[@class='subList']") ?? Enumerable.Empty<HtmlNode>();
		}

		/// <summary>
		/// Returns whether a subtitle node should be filtered from results.
		/// </summary>
		/// <param name="subtitleNode">The subtitle node to check.</param>
		/// <returns>Always <c>false</c>.</returns>
		protected override bool IsFiltered(HtmlNode subtitleNode) {
			return false;
		}

		/// <summary>
		/// Initializes a new instance of <see cref="MovieSubtitleSearcher"/>.
		/// </summary>
		/// <param name="query">The query associated with the search.</param>
		/// <param name="timeout">The timeout of the operation.</param>
		internal MovieSubtitleSearcher(ImdbSearchQuery query, TimeSpan timeout)
			: base(query, timeout) {
		}
	}

}