using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using SevenZip;
using SubtitleDownloader.Core;

namespace GammaJul.MediaPortal.Subtitles.SousTitresEu {

	/// <summary>
	/// Implementation of <see cref="ISubtitleDownloader"/> for http://www.sous-titres.eu/.
	/// </summary>
	public class SousTitresEuSubtitleDownloader : ISubtitleDownloader {

		/// <summary>
		/// Gets or sets the search timeout.
		/// </summary>
		public int SearchTimeout { get; set; }

		/// <summary>
		/// Gets the name of the subtitle downloader.
		/// </summary>
		/// <returns>Always returns "Sous-titres.eu".</returns>
		public string GetName() {
			return "Sous-titres.eu";
		}


		#region Search

		/// <summary>
		/// Returns whether a query has at least one supported language.
		/// </summary>
		/// <param name="query">The search query.</param>
		/// <returns><c>true</c> if <paramref name="query"/> as at least one supported language, <c>false</c> otherwise.</returns>
		/// <remarks>
		/// Sous-titres.eu only have english and french subtitles so we don't bother searching if we know in advance
		/// that none of these two languages are present in the query.
		/// </remarks>
		private static bool HasSupportedLanguage(SubtitleSearchQuery query) {
			return query.LanguageCodes.Any(code =>
				LanguageInfos.All.Any(info => info.IsoCode.Equals(code, StringComparison.OrdinalIgnoreCase)));
		}

		/// <summary>
		/// Searches for subtitles with a query string.
		/// </summary>
		/// <param name="query">The search query.</param>
		/// <returns>A list of subtitles matching <paramref name="query"/>.</returns>
		public List<Subtitle> SearchSubtitles(SearchQuery query) {
			return new List<Subtitle>(0);
		}

		/// <summary>
		/// Searches for serie subtitles with a serie name, season number and episode number.
		/// </summary>
		/// <param name="query">The search query.</param>
		/// <returns>A list of subtitles matching <paramref name="query"/>.</returns>
		public List<Subtitle> SearchSubtitles(EpisodeSearchQuery query) {
			if (query == null
			|| String.IsNullOrEmpty(query.SerieTitle)
			|| !HasSupportedLanguage(query))
				return new List<Subtitle>(0);

			var subtitleFetcher = new EpisodeSubtitleSearcher(query, TimeSpan.FromSeconds(SearchTimeout));
			return subtitleFetcher.FetchSubtitles();
		}

		/// <summary>
		/// Searches for movie subtitles with an IMDb ID.
		/// </summary>
		/// <param name="query">The search query.</param>
		/// <returns>A list of subtitles matching <paramref name="query"/>.</returns>
		public List<Subtitle> SearchSubtitles(ImdbSearchQuery query) {
			if (query == null
			|| String.IsNullOrEmpty(query.ImdbId)
			|| !HasSupportedLanguage(query))
				return new List<Subtitle>(0);

			var subtitleFetcher = new MovieSubtitleSearcher(query, TimeSpan.FromSeconds(SearchTimeout));
			return subtitleFetcher.FetchSubtitles();
		}

		#endregion


		#region Save

		/// <summary>
		/// Given a list of subtitles file names within a zip, returns a filtered set of file names for a given language code.
		/// </summary>
		/// <param name="archiveFileNames">A list of subtitles file names.</param>
		/// <param name="languageCode">The language code to use as a filter.</param>
		/// <returns>
		/// A <see cref="HashSet{T}"/> containing file names for the language identified with <paramref name="languageCode"/>,
		/// as well as file names not specific to a language. If there are no such files, all subtitles for other languages are returned.
		/// </returns>
		private static HashSet<string> GetFilesToExtract(IEnumerable<string> archiveFileNames, string languageCode) {
			var languageInfo = LanguageInfos.All.FirstOrDefault(info => String.Equals(info.IsoCode, languageCode, StringComparison.OrdinalIgnoreCase));
			var finalFileNames = new HashSet<string>();
			var wrongLanguageFileNames = new HashSet<string>();
			
			foreach (string fileName in archiveFileNames) {
				if (!fileName.EndsWith(".srt", StringComparison.OrdinalIgnoreCase)
				&& !fileName.EndsWith(".ass", StringComparison.OrdinalIgnoreCase))
					continue;

				bool wrongLanguageFound = LanguageInfos.All.Any(info => info != languageInfo && MatchLanguage(fileName, info));
				(wrongLanguageFound ? wrongLanguageFileNames : finalFileNames).Add(fileName);
			}

			return finalFileNames.Count > 0 ? finalFileNames : wrongLanguageFileNames;
		}

		/// <summary>
		/// Checks that a file name matches a given language.
		/// </summary>
		/// <param name="fileName">A file name.</param>
		/// <param name="languageInfo">A <see cref="LanguageInfo"/> used to check the file name.</param>
		/// <returns><c>true</c> if <paramref name="fileName"/> contains information related to <paramref name="languageInfo"/>.</returns>
		private static bool MatchLanguage(string fileName, LanguageInfo languageInfo) {
			return fileName.IndexOf("." + languageInfo.TwoLettersCode + ".", StringComparison.OrdinalIgnoreCase) >= 0
				|| fileName.StartsWith(languageInfo.VersionAbbreviation, StringComparison.OrdinalIgnoreCase)
				|| fileName.IndexOf("." + languageInfo.VersionAbbreviation + ".", StringComparison.OrdinalIgnoreCase) >= 0;
		}

		/// <summary>
		/// Extracts subtitles of a given language from a given zip file name.
		/// </summary>
		/// <param name="zipFileName">The zip file name that will be extracted.</param>
		/// <param name="languageCode">The language code of subtitles to extract.</param>
		/// <returns>A list of <see cref="FileInfo"/> representing extracted subtitles.</returns>
		private static List<FileInfo> ExtractSubtitles(string zipFileName, string languageCode) {
			string tempDirectory = Path.Combine(Path.GetTempPath(), Path.GetRandomFileName());
			Directory.CreateDirectory(tempDirectory);

			using (var extractor = new SevenZipExtractor(zipFileName)) {
				var filesToExtract = GetFilesToExtract(extractor.ArchiveFileNames, languageCode);
				extractor.ExtractFiles(args => {
					args.ExtractToStream = null;
					args.ExtractToFile = filesToExtract.Contains(args.ArchiveFileInfo.FileName)
						? Path.Combine(tempDirectory, FlattenFileName(args.ArchiveFileInfo.FileName))
						: null;
				});
			}

			return Directory.GetFiles(tempDirectory)
				.Select(fileName => new FileInfo(fileName))
				.ToList();
		}

		/// <summary>
		/// Flattens directories from an archive file name into a file name without hierarchy.
		/// </summary>
		/// <param name="fileName">The file name to flatten.</param>
		/// <returns>A flattened file name, or <paramref name="fileName"/> if it's already flat.</returns>
		/// <example>
		/// Passing "Serie.1x01.srt" will return "Series.1x01.srt".
		/// Passing "VF\Serie.1x01.srt" will return "[VF]Series.1x01.srt".
		/// </example>
		private static string FlattenFileName(string fileName) {
			string[] parts = fileName.Split(Path.DirectorySeparatorChar, Path.AltDirectorySeparatorChar);
			int length = parts.Length;

			if (length == 1)
				return parts[0];
			
			var builder = new StringBuilder();
			for (int i = 0; i < length; ++i) {
				string part = parts[i];
				if (i < length - 1)
					builder.Append('[');
				builder.Append(part);
				if (i < length - 1)
					builder.Append(']');
			}
			return builder.ToString();
		}

		/// <summary>
		/// Downloads, extracts and saves a subtitle.
		/// </summary>
		/// <param name="subtitle">The subtitle to save.</param>
		/// <returns>A list of <see cref="FileInfo"/> corresponding to several subtitles.</returns>
		public List<FileInfo> SaveSubtitle(Subtitle subtitle) {
			string tempFileName = Path.GetTempFileName() + ".zip";
			using (var webClient = new WebClient())
				webClient.DownloadFile(subtitle.Id, tempFileName);
			return ExtractSubtitles(tempFileName, subtitle.LanguageCode);
		}

		#endregion


	}

}