using System;
using System.Collections.Generic;
using System.Linq;
using HtmlAgilityPack;
using SubtitleDownloader.Core;

namespace GammaJul.MediaPortal.Subtitles.SousTitresEu {

	/// <summary>
	/// Base class for episode and movie search in sous-titres.eu.
	/// </summary>
	/// <typeparam name="TQuery">The type of query associated with the search.</typeparam>
	internal abstract class SubtitleSearcher<TQuery>
	where TQuery : SubtitleSearchQuery {
		private readonly TQuery _query;
		private TimeSpan _timeout;
		private string _title;

		internal const string SiteUrl = "http://www.sous-titres.eu/";

		/// <summary>
		/// Gets the query associated with the search.
		/// </summary>
		protected TQuery Query {
			get { return _query; }
		}

		/// <summary>
		/// Gets the URL of the page containing the listing of series or movies.
		/// </summary>
		protected abstract string ListPageUrl { get; }

		/// <summary>
		/// Gets the download base URL for links found onto <see cref="ListPageUrl"/>.
		/// </summary>
		protected abstract string DownloadBaseUrl { get; }

		/// <summary>
		/// Returns the title of the serie of movie queried.
		/// </summary>
		/// <returns>A title.</returns>
		protected abstract string GetTitleCore();

		/// <summary>
		/// Returns the title of the serie of movie queried. Caches the result.
		/// </summary>
		/// <returns>A title.</returns>
		private string GetTitle() {
			return _title ?? (_title = GetTitleCore());
		}

		/// <summary>
		/// Loads a <see cref="HtmlDocument"/> from a given URL.
		/// </summary>
		/// <param name="url">The URL to load the document from.</param>
		/// <returns>An instance of <see cref="HtmlDocument"/> representing the page at <paramref name="url"/>.</returns>
		protected HtmlDocument LoadDocument(string url) {
			var htmlWeb = new HtmlWeb {
				PreRequest = request => {
					request.Timeout = (int) _timeout.TotalMilliseconds;
					return true;
				}
			};
			HtmlDocument document = htmlWeb.Load(url);
			_timeout -= TimeSpan.FromMilliseconds(htmlWeb.RequestDuration);
			if (_timeout < TimeSpan.Zero)
				_timeout = TimeSpan.Zero;
			return document;
		}

		/// <summary>
		/// Fetches subtitles from sous-titres.eu.
		/// </summary>
		/// <returns>A list of <see cref="Subtitle"/>.</returns>
		internal List<Subtitle> FetchSubtitles() {
			var subtitles = new List<Subtitle>();
			string title = GetTitle();
			if (!String.IsNullOrEmpty(title)) {
				string serieUrl = GetItemUrl(GetTitle());
				if (serieUrl != null)
					subtitles.AddRange(GetSubtitles(serieUrl));
			}
			return subtitles;
		}

		/// <summary>
		/// Gets the first text of a HTML node.
		/// </summary>
		/// <param name="containerNode">The HTML node whose text will be read.</param>
		/// <returns>A title, or <c>null</c>.</returns>
		/// <remarks>This is different from <see cref="HtmlNode.InnerText"/> since only the first top-level piece of text will be retrieved.</remarks>
		protected static string GetFirstText(HtmlNode containerNode) {
			string title = null;
			HtmlNode textNode = containerNode.ChildNodes.FirstOrDefault(node => node.NodeType == HtmlNodeType.Text);
			if (textNode != null) {
				title = textNode.InnerText;
				if (title != null)
					title = title.Trim();
			}
			return title;
		}

		/// <summary>
		/// Retrieves a serie or movie page URL from the listing page.
		/// </summary>
		/// <param name="title">The serie or movie title.</param>
		/// <returns>An URL for <paramref name="title"/>, or <c>null</c>.</returns>
		private string GetItemUrl(string title) {
			HtmlDocument document = LoadDocument(ListPageUrl);
			HtmlNodeCollection linkNodes = document.DocumentNode.SelectNodes("//div[@class='items']//a");
			string url = null;
			if (linkNodes != null) {
				url = GetItemUrl(linkNodes, s => String.Equals(s, title, StringComparison.OrdinalIgnoreCase));
				if (String.IsNullOrEmpty(url)) {
					string titleWithoutYear = TitleWithoutYear(title);
					if (titleWithoutYear != title)
						url = GetItemUrl(linkNodes, s => String.Equals(TitleWithoutYear(s), titleWithoutYear, StringComparison.OrdinalIgnoreCase));
				}
			}
			return url;
		}

		/// <summary>
		/// Retrieves a serie or movie page URL from listing page nodes.
		/// </summary>
		/// <param name="linkNodes">The nodes containing titles.</param>
		/// <param name="titleMatcher">A function indicating if a title matches the searched title.</param>
		/// <returns>An URL, or <c>null</c>.</returns>
		private static string GetItemUrl(IEnumerable<HtmlNode> linkNodes, Func<string, bool> titleMatcher) {
			string url = null;
			HtmlNode serieNode = linkNodes.FirstOrDefault(node => titleMatcher(GetFirstText(node)));
			if (serieNode != null) {
				url = serieNode.GetAttributeValue("href", null);
				if (!String.IsNullOrEmpty(url)) {
					url = url.Trim();
					if (!String.IsNullOrEmpty(url))
						url = SiteUrl + url.Trim();
				}
			}
			return url;
		}

		/// <summary>
		/// Gets a title without the air year.
		/// </summary>
		/// <param name="title">The title whose year will be stripped.</param>
		/// <returns>If title had a year, returns <paramref name="title"/> without it; otherwise returns <paramref name="title"/>.</returns>
		/// <example>
		/// Passing "Merlin (2008)" will return "Merlin".
		/// Passing "Caprica" will return "Caprica".
		/// </example>
		private static string TitleWithoutYear(string title) {
			int length;
			if (!String.IsNullOrEmpty(title)
			&& (length = title.Length) > 6
			&& title[length - 6] == '('
			&& Char.IsDigit(title[length - 5])
			&& Char.IsDigit(title[length - 4])
			&& Char.IsDigit(title[length - 3])
			&& Char.IsDigit(title[length - 2])
			&& title[length - 1] == ')') {
				title = title.Substring(0, title.Length - 6).Trim();
			}
			return title;
		}

		/// <summary>
		/// Returns subtitles nodes potentially matching the query from a serie or movie page.
		/// </summary>
		/// <param name="documentNode">The document node of the serie or movie page.</param>
		/// <returns>An enumeration of <see cref="HtmlNode"/> containing potential subtitles.</returns>
		protected abstract IEnumerable<HtmlNode> GetSubtitleNodes(HtmlNode documentNode);

		private IEnumerable<Subtitle> GetSubtitles(string serieUrl) {
			HtmlDocument document = LoadDocument(serieUrl);

			var subtitlesNodes = GetSubtitleNodes(document.DocumentNode);
			return subtitlesNodes != null
				? subtitlesNodes.SelectMany(GetLanguageSubtitles)
				: Enumerable.Empty<Subtitle>();
		}

		/// <summary>
		/// Returns whether a subtitle node should be filtered from results.
		/// </summary>
		/// <param name="subtitleNode">The subtitle node to check.</param>
		/// <returns><c>true</c> if <paramref name="subtitleNode"/> should be filtered from results, <c>false</c> otherwise.</returns>
		protected abstract bool IsFiltered(HtmlNode subtitleNode);

		/// <summary>
		/// Gets all language subtitles from a subtitle node.
		/// </summary>
		/// <param name="subtitleNode">The subtitle node to use.</param>
		/// <returns>An enumeration of <see cref="HtmlNode"/> containing subtitles matching the query languages.</returns>
		private IEnumerable<Subtitle> GetLanguageSubtitles(HtmlNode subtitleNode) {
			if (IsFiltered(subtitleNode))
				yield break;

			string fileName = GetFirstText(subtitleNode.SelectSingleNode("span[@class='filenameSerie']"));
			if (String.IsNullOrEmpty(fileName))
				yield break;

			string zipUrl = subtitleNode.GetAttributeValue("href", null);
			if (String.IsNullOrEmpty(zipUrl))
				yield break;
			
			foreach (string languageCode in GetLanguageCodes(subtitleNode.SelectNodes("span[@class='lang']/img"))) {
				yield return new Subtitle(DownloadBaseUrl + zipUrl, GetTitle(), fileName, languageCode);
			}
		}

		/// <summary>
		/// Gets all languages for a given subtitle file, from flag images, filtered with query languages.
		/// </summary>
		/// <param name="imageNodes">A list of flag image nodes.</param>
		/// <returns>An enumeration of supported language codes retrived from the image tags.</returns>
		private IEnumerable<string> GetLanguageCodes(IEnumerable<HtmlNode> imageNodes) {
			foreach (HtmlNode imageNode in imageNodes) {
				string language = imageNode.GetAttributeValue("title", null);
				foreach (LanguageInfo languageInfo in LanguageInfos.All) {
					if (languageInfo.TwoLettersCode.Equals(language, StringComparison.OrdinalIgnoreCase)
					&& Query.HasLanguageCode(languageInfo.IsoCode))
						yield return languageInfo.IsoCode;
				}
			}
		}

		/// <summary>
		/// Initializes a new instance of <see cref="SubtitleSearcher{TQuery}"/>.
		/// </summary>
		/// <param name="query">The query associated with the search.</param>
		/// <param name="timeout">The timeout of the operation.</param>
		protected SubtitleSearcher(TQuery query, TimeSpan timeout) {
			_query = query;
			_timeout = timeout;
		}
	}

}